/*
 * Copyright 2013 Andre Simon <andre.simon1@gmx.de>
 * Copyright 2011 Benjamin Tissoires <benjamin.tissoires@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  ee the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/input.h>
#include <linux/hidraw.h>

#include <libudev.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

#define USB_VENDOR_ID_LOGITECH                  (__u32)0x046d
#define USB_DEVICE_ID_UNIFYING_RECEIVER         (__s16)0xc52b
#define USB_DEVICE_ID_UNIFYING_RECEIVER_2       (__s16)0xc532

#define VERSION "0.1"

// http://www.signal11.us/oss/udev/

int test_usb_device(const char* p){

      int usb_fd = open(p, O_RDWR|O_NONBLOCK);

      if (usb_fd < 0) {
	      perror("Unable to open device");
	      return usb_fd;
      }

      struct hidraw_devinfo info;
      /* Get Raw Info */
      int res = ioctl(usb_fd, HIDIOCGRAWINFO, &info);
      if (res < 0) {
	      printf("ERROR while getting info from device");
	      close(usb_fd);
	      return -1;
      } else {
	  if (info.bustype != BUS_USB ||
	      info.vendor != USB_VENDOR_ID_LOGITECH ||
	      (info.product != USB_DEVICE_ID_UNIFYING_RECEIVER &&
		info.product != USB_DEVICE_ID_UNIFYING_RECEIVER_2)) {
		 // errno = EPERM;
		//  perror("The given device is not a Logitech "
		//	  "Unifying Receiver\n");
		  close(usb_fd);
		  return -2;
	  }
      }

      char buf[256]={0};
      res = ioctl(usb_fd, HIDIOCGRAWNAME(256), buf);
      if (res>=0 ) {
	if (strncmp(buf, "Logitech USB Receiver", sizeof(buf))) {
	 // printf("Not a receiver...\n");
	  close(usb_fd);
	  return -3;
	}
      }

      return usb_fd;
}

int test_hidraw_files(){

        int hid_fd=0;
        struct udev *udev;
	struct udev_enumerate *enumerate;
	struct udev_list_entry *devices, *dev_list_entry;
	struct udev_device *dev;

	udev = udev_new();
	if (!udev) {
		printf("ERROR cannot initialize udev\n");
		return -1;
	}

	/* Create a list of the devices in the 'hidraw' subsystem. */
	enumerate = udev_enumerate_new(udev);
	udev_enumerate_add_match_subsystem(enumerate, "hidraw");
	udev_enumerate_scan_devices(enumerate);
	devices = udev_enumerate_get_list_entry(enumerate);

	udev_list_entry_foreach(dev_list_entry, devices) {
		const char *path;
		const char *nodePath;

		path = udev_list_entry_get_name(dev_list_entry);
		dev = udev_device_new_from_syspath(udev, path);

		nodePath=udev_device_get_devnode(dev);
		//printf("Device Node Path: %s\n", nodePath);
		if ( (hid_fd=test_usb_device(nodePath))>0) {
		  printf("Found USB Receiver: %s\n",(char*)nodePath);
		  udev_device_unref(dev);
		  break;
		}

		udev_device_unref(dev);
	}
	/* Free the enumerator object */
	udev_enumerate_unref(enumerate);
	udev_unref(udev);
	return hid_fd;
}

int wait_for_added_device(){

	struct udev *udev;
	udev = udev_new();
	if (!udev) {
		printf("ERROR cannot initialize udev\n");
		return -1;
	}
 	/* Set up a monitor to monitor hidraw devices */
	struct udev_monitor *mon = udev_monitor_new_from_netlink(udev, "udev");
	udev_monitor_filter_add_match_subsystem_devtype(mon, "hidraw", NULL);
	udev_monitor_enable_receiving(mon);
	/* Get the file descriptor (fd) for the monitor.
	   This fd will get passed to select() */
	int fd = udev_monitor_get_fd(mon);
	printf("Waiting - switch your device ON or press CTRL+C to exit.\n");

	while (1) {
		/* Set up the call to select(). In this case, select() will
		   only operate on a single file descriptor, the one
		   associated with our udev_monitor. Note that the timeval
		   object is set to 0, which will cause select() to not
		   block. */
		fd_set fds;
		struct timeval tv;
		int ret;

		FD_ZERO(&fds);
		FD_SET(fd, &fds);
		tv.tv_sec = 0;
		tv.tv_usec = 0;

		ret = select(fd+1, &fds, NULL, NULL, &tv);

		/* Check if our file descriptor has received data. */
		if (ret > 0 && FD_ISSET(fd, &fds)) {

			/* Make the call to receive the device.
			   select() ensured that this will not block. */
			struct udev_device *dev = udev_monitor_receive_device(mon);
			if (dev  && !strncmp(udev_device_get_action(dev),"add", 3)) {
				printf("\nDetected new device\n");
				printf("Node: %s\n", udev_device_get_devnode(dev));
				printf("Subsystem: %s\n", udev_device_get_subsystem(dev));
				udev_device_unref(dev);

				break;
			}
		}
		usleep(500*1000);
		printf(".");
		fflush(stdout);
	}
	udev_unref(udev);
	return 0;
}

int main(int argc, char **argv)
{
      int fd;

      char magic_sequence[] = {0x10, 0xFF, 0x80, 0xB2, 0x01, 0x00, 0x00};

      if (argc == 1) {

	printf("Searching for the Logitech Receiver...\n");
	if ((fd=test_hidraw_files())<=0){
	  printf("ERROR no device found - please connect the USB receiver and try again.\n");
	  return 1;
	}
      } else {
	if  (!strcmp(argv[1], "--version") || !strcmp(argv[1], "-V")){
	    printf("unipair version " VERSION "\n");
	    printf("Copyright 2011 Benjamin Tissoires\n");
	    printf("Copyright 2013 Andre Simon\n");
	    return 0;
	}  else if  (!strcmp(argv[1], "--help") || !strcmp(argv[1], "-h")){
	    printf("USAGE: unipair [OPTIONS] [FILE]\n");
	    printf("Run unipair without arguments to let it automatically search the receiver.\n");
	    return 0;
	}
	printf("Assuming that  %s is a Logitech Receiver...\n", argv[1]);
	if ((fd=test_usb_device(argv[1]))<=0){
	      errno = EINVAL;
	      return 1;
	}
      }

      printf("Switch your device OFF and press ENTER to continue.\n");
      getchar();//putchar (x);

      // Send the magic sequence to the Device
      int res = write(fd, magic_sequence, sizeof(magic_sequence));
      if (res < 0) {
	printf("ERROR: %d\n", errno);
      } else if (res == sizeof(magic_sequence)) {
	 wait_for_added_device();
      } else {
	      errno = ENOMEM;
	      printf("ERROR: %d were written instead of %ld.\n", res,
		      sizeof(magic_sequence));
      }

      close(fd);
      return 0;
}
